#include <stdio.h>
#include <string.h>

#define F(X,Y,Z) ((X) & (Y) | ~(X) & (Z))
#define G(X,Y,Z) ((X) & (Z) | (Y) & ~(Z))
#define H(X,Y,Z) ((X) ^ (Y) ^ (Z))
#define I(X,Y,Z) ((Y) ^ ((X) | ~(Z)))

#define bitRotL(w, s) (w << s | w >> 32 - s)

#define R(f,a,b,c,d,k,s,t) (a = b + bitRotL(a + f(b,c,d) + k + t, s))

typedef unsigned char byte;
typedef unsigned long int dword;

dword reverseWord(dword a) {
return (a << 24) & 0xFF000000 | (a << 8) & 0x00FF0000 | (a >> 8) & 0x0000FF00 | (a >> 24) & 0x000000FF;
}

int main(int argc, char * argv[]) {
// Checking parameters
int i;

// cmp first says if there is a digest to compare but later stores the return value
int cmp = 0; 
// file determines whether the parameters are correct or not
int file = 0;

for (i=1;i<argc;i++) {
	if (!memcmp(argv[i], "-c", 2)) cmp = i;
	else file = i;
}

if (cmp) {
	if (strlen (argv[cmp]) != 34) file = 0;
}

if (!file) {
	fprintf(stderr, "Incorrect parameters.\n");
	return 0;
}

// Loading file into memory
FILE * inFile = fopen(argv[file], "rb");
if (inFile == NULL) {
	fprintf(stderr, "Error opening file.\n");
	return 0;
}

fseek(inFile, 0, SEEK_END);
dword inSize = ftell(inFile);
fseek(inFile, 0, SEEK_SET);

/* MD5 Steps 1 and 2 are clumped together with step 4 to conserve memory */

// MD5 Step 3: Initializing MD Buffer
register dword A = 0x67452301;
register dword B = 0xefcdab89;
register dword C = 0x98badcfe;
register dword D = 0x10325476;

// MD5 Step 4: Processing in 16-Word Blocks
bool end = false;

while (!end) {
dword X[16] = {0};

if (fread(X, 4, 16, inFile) < 64 && feof(inFile)) {
	*((byte*)X + inSize % 64) = 0x80;
	
	X[14] = inSize << 3;
	X[15] = inSize >> 29;
	
	end = true;
}

dword AA = A;
dword BB = B;
dword CC = C;
dword DD = D;

// Round 1
R(F, A, B, C, D, X[0], 7, 0xd76aa478);
R(F, D, A, B, C, X[1], 12, 0xe8c7b756);
R(F, C, D, A, B, X[2], 17, 0x242070db);
R(F, B, C, D, A, X[3], 22, 0xc1bdceee);

R(F, A, B, C, D, X[4], 7, 0xf57c0faf);
R(F, D, A, B, C, X[5], 12, 0x4787c62a);
R(F, C, D, A, B, X[6], 17, 0xa8304613);
R(F, B, C, D, A, X[7], 22, 0xfd469501);

R(F, A, B, C, D, X[8], 7, 0x698098d8);
R(F, D, A, B, C, X[9], 12, 0x8b44f7af);
R(F, C, D, A, B, X[10], 17, 0xffff5bb1);
R(F, B, C, D, A, X[11], 22, 0x895cd7be);

R(F, A, B, C, D, X[12], 7, 0x6b901122);
R(F, D, A, B, C, X[13], 12, 0xfd987193);
R(F, C, D, A, B, X[14], 17, 0xa679438e);
R(F, B, C, D, A, X[15], 22, 0x49b40821);

// Round 2
R(G, A, B, C, D, X[1], 5, 0xf61e2562);
R(G, D, A, B, C, X[6], 9, 0xc040b340);
R(G, C, D, A, B, X[11], 14, 0x265e5a51);
R(G, B, C, D, A, X[0], 20, 0xe9b6c7aa);

R(G, A, B, C, D, X[5], 5, 0xd62f105d);
R(G, D, A, B, C, X[10], 9, 0x02441453);
R(G, C, D, A, B, X[15], 14, 0xd8a1e681);
R(G, B, C, D, A, X[4], 20, 0xe7d3fbc8);

R(G, A, B, C, D, X[9], 5, 0x21e1cde6);
R(G, D, A, B, C, X[14], 9, 0xc33707d6);
R(G, C, D, A, B, X[3], 14, 0xf4d50d87);
R(G, B, C, D, A, X[8], 20, 0x455a14ed);

R(G, A, B, C, D, X[13], 5, 0xa9e3e905);
R(G, D, A, B, C, X[2], 9, 0xfcefa3f8);
R(G, C, D, A, B, X[7], 14, 0x676f02d9);
R(G, B, C, D, A, X[12], 20, 0x8d2a4c8a);

// Round 3
R(H, A, B, C, D, X[5], 4, 0xfffa3942);
R(H, D, A, B, C, X[8], 11, 0x8771f681);
R(H, C, D, A, B, X[11], 16, 0x6d9d6122);
R(H, B, C, D, A, X[14], 23, 0xfde5380c);

R(H, A, B, C, D, X[1], 4, 0xa4beea44);
R(H, D, A, B, C, X[4], 11, 0x4bdecfa9);
R(H, C, D, A, B, X[7], 16, 0xf6bb4b60);
R(H, B, C, D, A, X[10], 23, 0xbebfbc70);

R(H, A, B, C, D, X[13], 4, 0x289b7ec6);
R(H, D, A, B, C, X[0], 11, 0xeaa127fa);
R(H, C, D, A, B, X[3], 16, 0xd4ef3085);
R(H, B, C, D, A, X[6], 23, 0x04881d05);

R(H, A, B, C, D, X[9], 4, 0xd9d4d039);
R(H, D, A, B, C, X[12], 11, 0xe6db99e5);
R(H, C, D, A, B, X[15], 16, 0x1fa27cf8);
R(H, B, C, D, A, X[2], 23, 0xc4ac5665);

// Round 4
R(I, A, B, C, D, X[0], 6, 0xf4292244);
R(I, D, A, B, C, X[7], 10, 0x432aff97);
R(I, C, D, A, B, X[14], 15, 0xab9423a7);
R(I, B, C, D, A, X[5], 21, 0xfc93a039);

R(I, A, B, C, D, X[12], 6, 0x655b59c3);
R(I, D, A, B, C, X[3], 10, 0x8f0ccc92);
R(I, C, D, A, B, X[10], 15, 0xffeff47d);
R(I, B, C, D, A, X[1], 21, 0x85845dd1);

R(I, A, B, C, D, X[8], 6, 0x6fa87e4f);
R(I, D, A, B, C, X[15], 10, 0xfe2ce6e0);
R(I, C, D, A, B, X[6], 15, 0xa3014314);
R(I, B, C, D, A, X[13], 21, 0x4e0811a1);

R(I, A, B, C, D, X[4], 6, 0xf7537e82);
R(I, D, A, B, C, X[11], 10, 0xbd3af235);
R(I, C, D, A, B, X[2], 15, 0x2ad7d2bb);
R(I, B, C, D, A, X[9], 21, 0xeb86d391);

A += AA;
B += BB;
C += CC;
D += DD;
}

// MD5 Step 5: Output
A = reverseWord(A);
B = reverseWord(B);
C = reverseWord(C);
D = reverseWord(D);

if (cmp) {
	dword SA, SB, SC, SD;
	sscanf(argv[cmp], "-c%08x%08x%08x%08x", &SA, &SB, &SC, &SD);
	if (A == SA && B == SB && C == SC & D == SD) cmp = 0;
	else cmp = 1;
} else {
	printf("%08X%08X%08X%08X\n", A, B, C, D);
}

// Cleanup
fclose(inFile);

return cmp;
}
